"""
sqery.py: Helpers for working with databases

Copyright 2014-2015, Outernet Inc.
Some rights reserved.

This software is free software licensed under the terms of GPLv3. See COPYING
file that comes with the source code, or http://www.gnu.org/licenses/gpl.txt.
"""

from __future__ import print_function

import calendar
import contextlib
import datetime
import logging
import re
import sqlite3

from sqlize import (From, Where, Group, Order, Limit, Select, Update, Delete,
                    Insert, Replace, sqlin, sqlarray)
from pytz import utc

from .utils import basestring


SLASH = re.compile(r'\\')
SQLITE_DATE_TYPES = ('date', 'datetime', 'timestamp')
MAX_VARIABLE_NUMBER = 999


def from_utc_timestamp(timestamp):
    """Converts the passed-in unix UTC timestamp into a datetime object."""
    dt = datetime.datetime.utcfromtimestamp(float(timestamp))
    return dt.replace(tzinfo=utc)


def to_utc_timestamp(dt):
    """Converts the passed-in datetime object into a unix UTC timestamp."""
    if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
        msg = "Naive datetime object passed. It is assumed that it's in UTC."
        logging.warning(msg)
    return calendar.timegm(dt.timetuple())


sqlite3.register_adapter(datetime.datetime, to_utc_timestamp)
for date_type in SQLITE_DATE_TYPES:
    sqlite3.register_converter(date_type, from_utc_timestamp)


class Row(sqlite3.Row):
    """ sqlite.Row subclass that allows attribute access to items """
    def __getattr__(self, key):
        return self[key]

    def get(self, key, default=None):
        key = str(key)
        try:
            return self[key]
        except IndexError:
            return default

    def __contains__(self, key):
        return key in self.keys()


class Connection(object):
    """ Wrapper for sqlite3.Connection object """
    def __init__(self, path=':memory:',):
        self.path = path
        self.connect()

    def connect(self):
        self._conn = sqlite3.connect(self.path,
                                     detect_types=sqlite3.PARSE_DECLTYPES)
        self._conn.row_factory = Row

        # Allow manual transaction handling, see http://bit.ly/1C7E7EQ
        self._conn.isolation_level = None
        # More on WAL: https://www.sqlite.org/isolation.html
        # Requires SQLite >= 3.7.0
        cur = self._conn.cursor()
        cur.execute('PRAGMA journal_mode=WAL;')
        logging.debug('Connected to database {}'.format(self.path))

    def close(self):
        self._conn.commit()
        self._conn.close()

    def __getattr__(self, attr):
        conn = object.__getattribute__(self, '_conn')
        return getattr(conn, attr)

    def __setattr__(self, attr, value):
        if not hasattr(self, attr) or attr == '_conn':
            object.__setattr__(self, attr, value)
        else:
            setattr(self._conn, attr, value)

    def __repr__(self):
        return "<Connection path='%s'>" % self.path


class Database(object):

    # Provide access to query classes for easier access
    sqlin = staticmethod(sqlin)
    sqlarray = staticmethod(sqlarray)
    From = From
    Where = Where
    Group = Group
    Order = Order
    Limit = Limit
    Select = Select
    Update = Update
    Delete = Delete
    Insert = Insert
    Replace = Replace
    MAX_VARIABLE_NUMBER = MAX_VARIABLE_NUMBER

    def __init__(self, conn, debug=False):
        self.conn = conn
        self.debug = debug
        self._cursor = None

    def _convert_query(self, qry):
        """ Ensure any SQLExpression instances are serialized

        :param qry:     raw SQL string or SQLExpression instance
        :returns:       raw SQL string
        """
        if hasattr(qry, 'serialize'):
            qry = qry.serialize()
        assert isinstance(qry, basestring), 'Expected qry to be string'
        if self.debug:
            print('SQL:', qry)
        return qry

    def query(self, qry, *params, **kwparams):
        """ Perform a query

        Any positional arguments are converted to a list of arguments for the
        query, and are used to populate any '?' placeholders. The keyword
        arguments are converted to a mapping which provides values to ':name'
        placeholders. These do not apply to SQLExpression instances.

        :param qry:     raw SQL or SQLExpression instance
        :returns:       cursor object
        """
        qry = self._convert_query(qry)
        self.cursor.execute(qry, params or kwparams)
        return self.cursor

    def execute(self, qry, *args, **kwargs):
        qry = self._convert_query(qry)
        self.cursor.execute(qry, *args, **kwargs)

    def executemany(self, qry, *args, **kwargs):
        qry = self._convert_query(qry)
        self.cursor.executemany(qry, *args, **kwargs)

    def executescript(self, sql):
        self.cursor.executescript(sql)

    def commit(self):
        self.conn.commit()

    def rollback(self):
        self.conn.rollback()
        self.conn.commit()

    def refresh_table_stats(self):
        self.execute('ANALYZE sqlite_master;')

    def acquire_lock(self):
        self.execute('BEGIN EXCLUSIVE;')

    def close(self):
        self.conn.close()
        # the cached cursor object must be reset, otherwise after reconnecting
        # it would still try to use it, and would run into the closed db issue
        self._cursor = None

    def reconnect(self):
        self.conn.connect()

    @property
    def connection(self):
        return self.conn

    @property
    def cursor(self):
        if self._cursor is None:
            self._cursor = self.conn.cursor()
        return self._cursor

    @property
    def results(self):
        return self.cursor.fetchall()

    @property
    def result(self):
        return self.cursor.fetchone()

    @contextlib.contextmanager
    def transaction(self, silent=False):
        self.execute('BEGIN;')
        try:
            yield self.cursor
            self.commit()
        except Exception:
            self.rollback()
            if silent:
                return
            raise

    @classmethod
    def connect(cls, dbpath):
        return Connection(dbpath)

    def __repr__(self):
        return "<Database connection='%s'>" % self.conn.path


class DatabaseContainer(dict):

    def __init__(self, connections, debug=False):
        databases = dict((n, Database(c, debug=debug))
                         for n, c in connections.items())
        super(DatabaseContainer, self).__init__(databases)
        self.__dict__ = self



import os
from setuptools import setup, find_packages

import squery_lite as pkg


def read(fname):
    """ Return content of specified file """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


VERSION = pkg.__version__

setup(
    name='squery-lite',
    version=VERSION,
    license='BSD',
    packages=[pkg.__name__],
    include_package_data=True,
    long_description=read('README.rst'),
    install_requires=[
        'pytz',
        'sqlize',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Software Development :: Libraries',
        'Environment :: Web Environment',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



"""
migrations.py: functions for managing migrations

Migration versions are tracked using a major and a minor version number. The
major version number is used to group migrations together, essentially allowing
to reset them back to 0. It was introduced in the later stages of development
when the need for resetting an existing database to an empty state came up, but
with the requirement to drop all the existing migrations for it as well.

Copyright 2014-2015, Outernet Inc.
Some rights reserved.

This software is free software licensed under the terms of GPLv3. See COPYING
file that comes with the source code, or http://www.gnu.org/licenses/gpl.txt.
"""

import importlib
import logging
import os
import re
import sys


VERSION_MULTIPLIER = 10000
SQLITE_EXTENSIONS = ('sqlite', 'sqlite-shm', 'sqlite-wal')
GET_VERSION_SQL = "PRAGMA user_version"
SET_VERSION_SQL = "PRAGMA user_version = {version:d}"
PYMOD_RE = re.compile(r'^((\d{2})_(\d{2})_[^.]+)\.pyc?$', re.I)


def get_mods(package):
    """ List all loadable python modules in a directory

    This function looks inside the specified directory for all files that look
    like Python modules with a numeric prefix and returns them. It will omit
    any duplicates and return file names without extension.

    :param package: package object
    :returns:       list of tuples containing filename without extension,
                    major_version and minor_version
    """
    pkgdir = package.__path__[0]
    matches = filter(None, [PYMOD_RE.match(f) for f in os.listdir(pkgdir)])
    parse_match = lambda groups: (groups[0], int(groups[1]), int(groups[2]))
    return sorted(list(set([parse_match(m.groups()) for m in matches])),
                  key=lambda x: (x[1], x[2]))


def get_new(modules, min_major_version, min_minor_version):
    """ Get list of migrations that haven't been run yet

    :param modules:           iterable containing module names
    :param min_major_version: minimum major version
    :param min_minor_version: minimum minor version
    :returns:                 return an iterator that yields only items which
                              versions are >= min_ver
    """
    for mod_data in modules:
        (modname, mod_major_version, mod_minor_version) = mod_data
        if (mod_major_version > min_major_version or
                (mod_major_version == min_major_version and
                    mod_minor_version >= min_minor_version)):
            yield mod_data


def load_mod(module, package):
    """ Load a module named ``module`` from given search``path``

    The module path prefix is set according to the ``prefix`` argument.
    By defualt the module is loaded as if it comes from a global
    'db_migrations' package.  As such, it may conflict with any 'db_migration'
    package. The module can be looked up in ``sys.modules`` as
    ``db_migration.MODNAME`` where ``MODNAME`` is the name supplied as
    ``module`` argument. Keep in mind that relative imports from within the
    module depend on this prefix.

    This function raises an ``ImportError`` exception if module is not found.

    :param module:  name of the module to load
    :param package: package object
    :returns:       module object
    """
    name = '%s.%s' % (package.__name__, module)
    if name in sys.modules:
        return sys.modules[name]
    return importlib.import_module(name, package=package.__name__)


def drop_db(db):
    """ Delete the passed in database, and reconnect afterwards.

    :param db:  database object
    """
    path, extension = os.path.splitext(db.connection.path)
    db.close()
    for ext in SQLITE_EXTENSIONS:
        file_path = '.'.join([path, ext])
        if os.path.exists(file_path):
            os.remove(file_path)

    db.reconnect()


def pack_version(major_version, minor_version):
    """Pack the two version integers into one int."""
    return major_version * VERSION_MULTIPLIER + minor_version


def unpack_version(version):
    """Unpack a single version integer into the two major and minor
    components."""
    minor_version = version % VERSION_MULTIPLIER
    major_version = (version - minor_version) / VERSION_MULTIPLIER
    return (major_version, minor_version)


def get_version(db):
    """ Query database and return migration version. WARNING: side effecting
    function! if no version information can be found, any existing database
    matching the passed one's name will be deleted and recreated.

    :param db:  connetion object
    :returns:   current migration version
    """
    db.query(GET_VERSION_SQL)
    version = db.result.user_version
    if not version:
        drop_db(db)
        return (0, 0)

    return unpack_version(version)


def set_version(db, major_version, minor_version):
    """ Set database migration version

    :param db:             connetion object
    :param major_version:  integer major version of migration
    :param minor_version:  integer minor version of migration
    """
    version = pack_version(major_version, minor_version)
    db.query(SET_VERSION_SQL.format(version=version))


def run_migration(major_version, minor_version, db, mod, conf={}):
    """ Run migration script

    :param major_version: major version number of the migration
    :param minor_version: minor version number of the migration
    :param db:            database connection object
    :param path:          path of the migration script
    :param conf:          application configuration (if any)
    """
    with db.transaction():
        mod.up(db, conf)
        set_version(db, major_version, minor_version)


def migrate(db, package, conf={}):
    """ Run all migrations that have not been run

    Migrations will be run inside a transaction.

    :param db:              database connection object
    :param package:         package that contains the migrations
    :param conf:            application configuration object
    """
    (current_major_version, current_minor_version) = get_version(db)
    package = importlib.import_module(package)
    logging.debug('Migration version for %s is %s.%s',
                  package.__name__,
                  current_major_version,
                  current_minor_version)
    mods = get_mods(package)
    migrations = get_new(mods,
                         current_major_version,
                         current_minor_version + 1)
    for (modname, major_version, minor_version) in migrations:
        mod = load_mod(modname, package)
        run_migration(major_version, minor_version, db, mod, conf)
        logging.debug("Finished migrating to %s", modname)
    db.refresh_table_stats()
